""" implements the various intel endpoint things """

import json

from loguru import logger

from .utilities import validate_kwargs

__all__ = ['get_intel_indicators']
THIS_SHOULD_FAIL = "True"

def get_intel_indicators(self, **kwargs):
    """ Get info about indicators that match provided FQL filters.

    https://assets.falcon.crowdstrike.com/support/api/swagger.html#/intel/QueryIntelIndicatorEntities

- offset (int) - Set the starting row number to return indicators from.
    Defaults to 0.

- limit (int) - Set the number of indicators to return.
    The number must be between 1 and 50000

- sort (str) - Order fields in ascending or descending order.
    Ex: published_date|asc.

- filter (str) - Filter your query by specifying FQL filter parameters.
    Filter parameters include:
    _marker, actors, deleted, domain_types, id, indicator,
    ip_address_types, kill_chains, labels, labels.created_on,
    labels.last_valid_on, labels.name, last_updated, malicious_confidence, malware_families, published_date, reports, targets, threat_types, type, vulnerabilities.

- q (str) - Perform a generic substring search across all fields.
- include_deleted (bool) - If true, include both published and deleted indicators in the response.
    Defaults to false.
    """

    method = 'get'
    uri = '/intel/combined/indicators/v1'
    args_validation = {
        'offset' : int,
        'limit' : int,
        'sort' : str,
        'filter' : str,
        'query' : str,
        'include_deleted' : bool,
    }
    validate_kwargs(args_validation, kwargs)
    logger.debug(f"kwargs: {json.dumps(kwargs)}")

    if 'limit' in kwargs:
        if (kwargs.get("limit") < 1) or (kwargs.get("limit") > 50000):
            raise ValueError("limit needs to be from 1-50000")

    # TODO: it's possible this'll throw a 400 error with the following:
    #     {
    #     "meta": {
    #         "query_time": 0.000331244,
    #         "powered_by": "msa-api",
    #         "trace_id": "xxxxxxxxxxx"
    #     },
    #     "resources": [],
    #     "errors": [
    #         {
    #         "code": 400,
    #         "message": "restricted API can only start from offset=0"
    #         }
    #     ]
    #     }
    # This doesn't happen on all accounts but it should be handled with better logging

    response = self.request(uri=uri,
                            request_method=method,
                            data=kwargs,
                            )
    logger.debug(response.json())
    return response.json()



""" handles interactions with the "hosts" category """

import json

from loguru import logger

from .utilities import validate_kwargs

HOST_ACTION_NAMES = [
    'contain',
    'lift_containment',
    'hide_host',
    'unhide_host',
]

def host_action(self, **kwargs):
    """ Take various actions on the hosts in your environment.
        Contain or lift containment on a host.
        Delete or restore a host.

        action_name : str (required) one of the following
            - contain
            - lift_containment
            - hide_host
            - unhide_host
        ids: str (required)
            - ids should be a list of strings
    """
    args_validation = {
        'action_name' : str,
        'ids' : list,
    }
    validate_kwargs(args_validation, kwargs, required=args_validation.keys())
    if kwargs.get('action_name') not in HOST_ACTION_NAMES:
        error_message = f"Invalid action_name={kwargs.get('action_name')} valid options are {','.join(HOST_ACTION_NAMES)}"
        logger.error(error_message)
        raise ValueError(error_message)

    uri = '/devices/entities/devices-actions/v2'
    method = 'post'

    response = self.request(uri=uri,
                            request_method=method,
                            data=kwargs,
                            )
    logger.debug(f"Request body: {response.request.body}")
    return response.json()

def hosts_detail(self, **kwargs):
    """ Get details on one or more hosts by providing agent IDs (AID). You can get a host's agent IDs (AIDs) from the /devices/queries/devices/v1 endpoint, the Falcon console or the Streaming API

    arguments:

    - ids : list (list of Agent IDs: required)
    """
    uri = '/devices/entities/devices/v1'
    method = 'get'

    args_validation = {
        'ids' : list,
    }

    validate_kwargs(args_validation, kwargs, required=args_validation.keys())

    response = self.request(uri=uri,
                            request_method=method,
                            data=kwargs,
                            )
    return response.json()



def hosts_hidden(self, **kwargs):
    """ Retrieve hidden hosts that match the provided filter criteria.
    offset = The offset to start retrieving records from
    limit = The maximum records to return. [1-5000]
    sort = The property to sort by (e.g. status.desc or hostname.asc)
    filter = The filter expression that should be used to limit the results

    returns the json object, resources are a list of host IDs
    """

    args_validation = {
        'offset' : int,
        'limit' : int,
        'sort' : str,
        'filter' : str,
    }
    validate_kwargs(args_validation, kwargs)
    uri = '/devices/queries/devices-hidden/v1'
    method = 'get'

    response = self.request(uri=uri,
                            request_method=method,
                            data=kwargs,
                            )
    return response.json()


# GET
# /devices/queries/devices-scroll/v1
# Search for hosts in your environment by platform, hostname, IP, and other criteria with continuous pagination capability (based on offset pointer which expires after 2 minutes with no maximum limit)

def hosts_query_devices(self, **kwargs):
    """ Search for hosts in your environment by platform, hostname, IP, and other criteria.
    offset = The offset to start retrieving records from
    limit = The maximum records to return. [1-5000]
    sort = The property to sort by (e.g. status.desc or hostname.asc)
    filter = The filter expression that should be used to limit the results

    returns the json object, resources are a list of host IDs
    """

    args_validation = {
        'offset' : int,
        'limit' : int,
        'sort' : str,
        'filter' : str,
    }
    validate_kwargs(args_validation, kwargs)

    uri = '/devices/queries/devices/v1'
    method = 'get'
    logger.debug(json.dumps(kwargs, indent=2))
    response = self.request(uri=uri,
                            request_method=method,
                            data=kwargs,
                            )
    return response.json()



""" implements the various detects endpoint things """

from loguru import logger

VALID_DETECT_STATUS = ['new', 'in_progress', 'true_positive', 'false_positive', 'ignored']

def get_detects(self, **kwargs):
    """ Search for detection IDs that match a given query

    offset (int) (query) The first detection to return, where 0 is the latest detection. Use with the limit parameter to manage pagination of results.

    limit (int) (query)	The maximum number of detections to return in this response (default: 9999; max: 9999). Use with the offset parameter to manage pagination of results.

    sort (str) (query)	 Sort detections using these options:
        - first_behavior: Timestamp of the first behavior associated with this detection
        - last_behavior: Timestamp of the last behavior associated with this detection
        - max_severity: Highest severity of the behaviors associated with this detection
        - max_confidence: Highest confidence of the behaviors associated with this detection
        - adversary_id: ID of the adversary associated with this detection, if any
        - devices.hostname: Hostname of the host where this detection was detected
        Sort either asc (ascending) or desc (descending). For example: last_behavior|asc

    filter (str) (query) Filter detections using a query in Falcon Query Language (FQL) An asterisk wildcard * includes all results.
        Common filter options include:
        - status
        - device.device_id
        - max_severity
        The full list of valid filter options is extensive. Review it in the documentation inside the Falcon console.

    q (str) (query)	Search all detection metadata for the provided string
"""
    uri = '/detects/queries/detects/v1'
    response = self.request(uri=uri,
                            request_method='get',
                            data=kwargs,
                            )
    logger.debug(response)
    response.raise_for_status()

    return response.json()

def get_detections(self, ids: list):
    """ view detection information """

    if not isinstance(ids, list):
        raise TypeError(f"ids should be of type 'list', got '{type(ids)}'")

    uri = '/detects/entities/summaries/GET/v1'
    response = self.request(uri=uri,
                            request_method='post',
                            data={'ids' : ids},
                            )
    logger.debug(response)
    response.raise_for_status()

    return response.json()

def update_detection(self, **kwargs):
    """ modify the date, assignee and visibility of detections

    assigned_to_uuid (string) - uuid of user ID that the task is assigned to

    ids (list) - one or more detection ID - from get_detects()

    show_in_ui (bool) - show the detection in falcohn. Most commonly used together with the status key's false_positive value

    status (str) - one of the following:
     - new
     - in_progress
     - true_positive
     - false_positive
     - ignored

    comment (str) optional comment to add to the detection
    """
    uri = '/detects/entities/detects/v2'

    if 'status' in kwargs and kwargs.get('status') not in VALID_DETECT_STATUS:
        set_status = kwargs.get('status')
        raise ValueError(f"Status '{set_status}' invalid - should be in {VALID_DETECT_STATUS}")
    response = self.request(uri=uri,
                            request_method='patch',
                            data=kwargs,
                            )
    logger.debug(response)
    response.raise_for_status()

    return response.json()



""" handles the event-streams endpoints """


def get_event_streams(self, appid: str, format_str: str = "json"):
    """
    Discover all event streams in your environment
    https://assets.falcon.crowdstrike.com/support/api/swagger.html#/event-streams/listAvailableStreamsOAuth2
    """
    uri = '/sensors/entities/datafeed/v2'
    data = {
        'format' : format_str,
    }
    if appid:
        data['appId'] = appid
    else:
        raise ValueError("Need to specify an appid")
    response = self.request(
        uri=uri,
        request_method='get',
        data=data,
        )
    return response.json()

def refresh_event_stream(self, partition: int, appid: str):
    """ refresh an active event stream. use the url returned from get_event_streams
    """

    raise NotImplementedError("sorry, not done yet")
    # action_name = 'refresh_active_stream_session'
    # example valid response
    # {
    # "errors": [
    #     {
    #     "code": 0,
    #     "id": "string",
    #     "message": "string"
    #     }
    # ],
    # "meta": {
    #     "pagination": {
    #     "limit": 0,
    #     "offset": 0,
    #     "total": 0
    #     },
    #     "powered_by": "string",
    #     "query_time": 0,
    #     "trace_id": "string",
    #     "writes": {
    #     "resources_affected": 0
    #     }
    # }
    # }



""" handles interactions with the "hosts" category """

from loguru import logger

from .utilities import validate_kwargs

def get_host_groups(self, **kwargs):
    """ Retrieve a set of host groups by specifying their IDs
    Returns a set of Host Groups which match the filter criteria

        ids: str (required)
            - ids should be a list of strings
    """
    args_validation = {
        'ids' : list,
    }
    validate_kwargs(args_validation, kwargs, required=args_validation.keys())

    uri = '/devices/entities/host-groups/v1'
    method = 'get'

    response = self.request(uri=uri,
                            request_method=method,
                            data=kwargs,
                            )
    logger.debug(f"Request body: {response.request.body}")
    return response.json()

def search_host_groups(self, **kwargs):
    """ Search for Host Groups in your environment by providing an FQL filter and paging details.
    Returns a set of Host Groups which match the filter criteria

    - filter (string) - The filter expression that should be used to limit the results

        *** FQL is case sensitive - ie it has to be in lower case ***

    - offset (integer) - The offset to start retrieving records from
    - limit (integer) - The maximum records to return. [1-5000]
    - sort (string) - field to sort by
    """
    args_validation = {
        'filter' : str,
        'offset' : int,
        'limit' : int,
        'sort' : str,

    }
    validate_kwargs(args_validation, kwargs)

    uri = '/devices/combined/host-groups/v1'
    method = 'get'

    response = self.request(uri=uri,
                            request_method=method,
                            data=kwargs,
                            )
    logger.debug(f"Request body: {response.request.body}")
    return response.json()

def update_host_group(self, **kwargs):
    """
    Update Host Groups by specifying the ID of the group and details to update

    - assignment_rule (string) - the FQL for matching entities
    - description (string) - group description
    - id (string) - the group id for matching, won't update it
    - name (string) - the group name
    """
    args_validation = {"assignment_rule": str,
                       "description": str,
                       "id": str,
                       "name": str,
                      }
    required_args = ['id']

    validate_kwargs(args_validation=args_validation, kwargs=kwargs, required=required_args)
    uri = '/devices/entities/host-groups/v1'
    method = 'patch'
    # build the data for the update
    update_data = {'resources' : [kwargs]}

    response = self.request(uri=uri,
                            request_method=method,
                            data=update_data,
                            )
    logger.debug(f"Request body: {response.request.body}")
    return response.json()

def create_host_group(self, **kwargs):
    """
    Create Host Groups by specifying details

    group_type should be static or dynamic

    - name (string) (required) - the group name
    - description (string) (required) - group description
    - group_type (string) (required) - the group type
    - assignment_rule (string) - the filter for matching entities - required for dynamic groups - eg "(machine_domain:'example.com')"

    """
    args_validation = {"assignment_rule": str,
                       "description": str,
                       "group_type": str,
                       "name": str,
                      }

    validate_kwargs(args_validation=args_validation, kwargs=kwargs, required=list(args_validation.keys()))

    uri = '/devices/entities/host-groups/v1'
    method = 'post'

    data = {
        'resources' : [kwargs]
    }

    response = self.request(uri=uri,
                            request_method=method,
                            data=data,
                            )
    logger.debug(f"Request body: {response.request.body}")
    return response.json()

def delete_host_groups(self, **kwargs):
    """ deletes host groups based on a list of IDs

    - ids (list) (required) - the IDs of the groups to delete
    """
    args_validation = {"ids": list,
                      }

    validate_kwargs(args_validation=args_validation, kwargs=kwargs, required=list(args_validation.keys()))

    uri = '/devices/entities/host-groups/v1'
    method = 'delete'

    response = self.request(uri=uri,
                            request_method=method,
                            data=kwargs,
                            )
    return response.json()



""" handler for incidents """

from loguru import logger

from .utilities import validate_kwargs

VALID_ACTION_KEYS = ['add_tag', 'delete_tag', 'update_name', 'update_description', 'update_status']

__all__ = [
    'incidents_get_crowdscores',
    'incidents_perform_actions',
    'incidents_get_details',
    'incidents_behaviors_by_id',
    'incidents_query_behaviors',
    'incidents_query',
    ]

ACTION_KEYS = ('name', 'value') # keys used in the body of the incidents_perform_action call

def incidents_get_crowdscores(self, **kwargs):
    """ Query environment wide CrowdScore and return the entity data """
    # uri = '/incidents/combined/crowdscores/v1'
    # method = 'get'
    raise NotImplementedError

def incidents_perform_actions(self, **kwargs):
    """ Perform a set of actions on one or more incidents, such as adding tags or comments or updating the incident name or description

    Documentation here: https://falcon.crowdstrike.com/support/documentation/86/detections-monitoring-apis#modify-incidents

    Valid Action_parameters
        add_tag - Adds the associated value as a new tag on all the incidents of the ids list.
        delete_tag - Deletes tags matching the value from all the incidents in the ids list
        update_name - Updates the name to the parameter value of all the incidents in the ids list.
        update_description - Updates the description to the parameter value of all the incidents listed in the ids.
        update_status - Updates the status to the parameter value of all the incidents in the ids list.
            Valid values for status are 20, 25, 30, 40: (also in crowdstrike.INCIDENT_STATUS_LOOKUP)
                20: New
                25: Reopened
                30: In Progress
                40: Closed
"""

    uri = '/incidents/entities/incident-actions/v1'
    method = 'post'
    args_validation = {
        'action_parameters' : list,
        'ids' : list,
    }
    validate_kwargs(args_validation, kwargs, required=args_validation.keys())

    # start validation of the action/value body
    for action in kwargs.get('action_parameters'):
        if not isinstance(action, dict):
            raise ValueError()
        if sorted(ACTION_KEYS) != sorted(set(action.keys())):
            raise ValueError(f"Keys for action_parameter have to be name, value only, got: {sorted(set(action.keys()))}")
        for key in ACTION_KEYS:
            if not isinstance(action.get(key), str):
                raise ValueError(f"Values for action_parameter have to be string, value only, {key} was {type(action.get(key))}")
            if key == 'name' and action.get(key) not in VALID_ACTION_KEYS:
                raise ValueError(f"Invalid action_parameter - set to '{action.get(key)}', not in {','.join(VALID_ACTION_KEYS)}")
    # end validation of the action/value body

    logger.debug(kwargs)
    response = self.request(uri=uri,
                            request_method=method,
                            data=kwargs,
                            )
    return response.json()


def incidents_get_details(self, **kwargs):
    """ Get details on incidents by providing a list of incident IDs

    returns the raw object so you can look for errors and pagination and so forth

    requires:
    - ids (list) - a list of incident IDs

    returns JSON data, response key has the following sub-keys: [
                        'incident_id', 'incident_type', 'cid',
                        'host_ids', 'hosts',
                        'created', 'start', 'end', 'state',
                        'status', 'tactics', 'techniques', 'objectives', 'users', 'fine_score',
                        ])

    swagger docs: https://assets.falcon.crowdstrike.com/support/api/swagger.html#/incidents/GetIncidents
    """
    uri = '/incidents/entities/incidents/GET/v1'
    method = 'post'
    args_validation = {
        'ids' : list,
    }
    validate_kwargs(args_validation, kwargs, required=args_validation.keys())
    response = self.request(uri=uri,
                            request_method=method,
                            data=kwargs,
                            )
    return response.json()

def incidents_behaviors_by_id(self, **kwargs):
    """Get details on behaviors by providing behavior IDs """
    # uri = '/incidents/entities/behaviors/GET/v1'
    # method = 'post'
    raise NotImplementedError

def incidents_query_behaviors(self, **kwargs):
    """Search for behaviors by providing an FQL filter, sorting, and paging details"""
    # uri = '/incidents/queries/behaviors/v1'
    # method = 'get'
    raise NotImplementedError

def incidents_query(self, **kwargs):
    """ Search for incidents by providing an FQL filter, sorting, and paging details

    returns a list of incidents in the format like "inc:aaaabbbbc9b94d0095fde66d407289ec:aaaabbbbe17048ad99f22746082617b5"

    docs: https://falcon.crowdstrike.com/support/documentation/86/detections-monitoring-apis

    args:
        - sort (str)
        - filter (str)
        - offset (int)
        - limit (int) - max 500, min 1

    filter examples:
        status: '20' # new
        status: '25' # reopened
        status: '30' # in progress
        status: '40' # closed

        score_range:'7.5 - 10'

        tags: 'True Positive', 'Ignored', 'Lateral Movement'

    """
    uri = '/incidents/queries/incidents/v1'
    method = 'get'
    args_validation = {
        'sort' : str,
        'filter' : str,
        'offset' : int,
        'limit' : int,
    }
    validate_kwargs(args_validation, kwargs)

    if 'limit' in kwargs:
        if int(kwargs.get('limit')) > 500:
            raise ValueError("Maximum of 500 for 'limit' on this endpoint.")
        if int(kwargs.get('limit')) < 1:
            raise ValueError("Minimum of 1 for 'limit' on this endpoint.")

    response = self.request(uri=uri,
                            request_method=method,
                            data=kwargs,
                            )
    if response.status_code == 400:
        logger.debug("Got 400 status code, potentially too large a response (max 500), or invalid filter.")
    if 'resources' in response.json():
        data = response.json().get('resources')
    else:
        logger.error("Didn't get a response")
        data = response.json()
    return data



""" implements the various sensor-download endpoint things """

import errno

from loguru import logger

def download_sensor(self, sensorid: str, destination_filename: str):
    """ downloads a sensor id to the filename """
    uri = '/sensors/entities/download-installer/v1'
    data = {
        'id' : sensorid,
    }
    try:
        logger.debug(f"Writing intaller to {destination_filename}")
        with open(destination_filename, 'wb') as file_handle:
            response = self.request(
                uri=uri,
                request_method='get',
                data=data,
            )
            logger.debug(response.headers)
            response.raise_for_status()

            file_handle.write(response.content)
        return True

    except IOError as file_write_error:
        error_number = file_write_error.errno
        error_string = file_write_error.strerror
        if error_number == errno.EACCES:
            logger.error(f"error {error_number}, {error_string} - Permission fail")
        elif error_number == errno.EISDIR:
            logger.error(f"error {error_number}, {error_string} - Path is a directory")
        else:
            logger.error(f"error {error_number}, {error_string}")
        return False

def get_ccid(self):
    """ returns the CCID for installers """
    req = self.request(request_method='get', uri="/sensors/queries/installers/ccid/v1")
    req.raise_for_status()

    if req.status_code == 200:
        retval = req.json().get('resources')[0]
    else:
        retval = False
    return retval

def get_latest_sensor_id(self, filter_string: str = ""):
    """ returns the ids of the latest sensor IDs

        suggested filter: 'platform:mac' or 'platform:windows'
    """
    response = self.get_sensor_installer_ids(
        sort_string="release_date|desc",
        filter_string=filter_string,)
    if response:
        retval = response[0]
    else:
        retval = False
    return retval

def get_sensor_installer_details(self, sensorid: str):
    """
    returns a dict about a particular sensor ID, or False if it can't find anything useful
    """
    logger.debug(f"Sensor ID: {sensorid}")
    uri = "/sensors/entities/installers/v1"
    data = {
        'ids' : sensorid,
        }
    response = self.request(uri=uri,
                            request_method='get',
                            data=data,
                            )

    response.raise_for_status()
    logger.debug(response.headers)
    if not response.json().get('resources', False):
        retval = False
    else:
        retval = response.json().get('resources', False)[0]
    return retval

def get_sensor_installer_ids(self, sort_string: str = "", filter_string: str = ""):
    """
    returns a list of installer IDs, they're a list of SHA256's
    """
    logger.debug(f"sort_string: '{sort_string}', filter_string: '{filter_string}'")
    uri = '/sensors/queries/installers/v1'

    data = {
        'sort' : sort_string,
        'filter' : filter_string,
    }

    response = self.request(request_method='get', uri=uri, data=data)
    #logger.debug(response.json())
    logger.debug("Request headers")
    logger.debug(response.request.headers)
    #logger.debug(dir(response))
    response.raise_for_status()


    # TODO: handle pagination
    return response.json().get('resources', False)



""" handles interactions with the "hosts" category """

import json
import time

from loguru import logger

from .utilities import validate_kwargs

def list_rtr_session_ids(self, **kwargs):
    """ Get a list of session_ids

    https://assets.falcon.crowdstrike.com/support/api/swagger.html#/real-time-response/RTR_ListAllSessions

    - offset (int) - starting index in list to query (ie, start from #5)
    - limit (int) - number of ids to return
    - sort (string) - sort by spec. Ex: 'date_created|asc’.
    - filter (string) - Optional filter criteria in the form of an FQL  query. “user_id” can accept a special value ‘@me’ which will restrict results to records with current user’s ID.
        example: filter="user_id: '@me'"

    """
    method = 'get'
    uri = '/real-time-response/queries/sessions/v1'
    args_validation = {
        'offset' : int,
        'limit' : int,
        'sort' : str,
        'filter' : str,
    }
    validate_kwargs(args_validation, kwargs)
    logger.debug(f"Querying RTR session list, kwargs: {json.dumps(kwargs)}")
    response = self.request(uri=uri,
                            request_method=method,
                            data=kwargs,
                            )
    logger.debug(response.json())
    return response.json()

def create_rtr_session(self, **kwargs):
    """ Create a new RTR session, will retrieve an existing session if exists.

    - device_id (string) - The host agent ID to initialize the RTR session on. RTR will retrieve an existing session for the calling user on this host
    """
    args_validation = {
        'device_id' : str,
    }
    validate_kwargs(args_validation, kwargs, required=args_validation.keys())

    uri = '/real-time-response/entities/sessions/v1'
    method = 'post'

    logger.debug(f"Creating RTR session for device_id={kwargs.get('device_id')}")
    response = self.request(uri=uri,
                            request_method=method,
                            data=kwargs,
                            )
    logger.debug(f"Request body: {response.request.body}")
    return response.json()

def delete_rtr_session(self, **kwargs):
    """ Delete a RTR session.

    - session_id (string) - RTR session ID
    """
    args_validation = {
        'session_id' : str,
    }
    validate_kwargs(args_validation, kwargs, required=args_validation.keys())

    uri = '/real-time-response/entities/sessions/v1'
    method = 'delete'

    response = self.request(uri=uri,
                            request_method=method,
                            data=kwargs,
                            )
    logger.debug(f"Request body: {response.request.body}")
    logger.debug(f"Response: {response.text}")
    return response

def rtr_execute_command(self, **kwargs):
    """ execute a command on a single host
    https://assets.falcon.crowdstrike.com/support/api/swagger.html#/real-time-response/RTR_ExecuteCommand

    - base_command (string) (required) - one of:
        cat
        cd
        clear
        env
        eventlog
        filehash
        getsid
        help
        history
        ipconfig
        ls
        mount
        netstat
        ps
        reg query
    - command_string (string) (required) - arguments to the command
    - session_id - (string) RTR session ID to run the command on

    returns a dict wtih
    """
    uri = '/real-time-response/entities/command/v1'
    method = 'post'
    args_validation = {
        'base_command' : str,
        'command_string' : str,
        'session_id' : str,
    }
    validate_kwargs(args_validation, kwargs, required=args_validation.keys())

    response = self.request(uri=uri,
                            request_method=method,
                            data=kwargs,
                            )
    logger.debug(f"Request body: {response.request.body}")
    logger.debug(f"Response: {response.text}")

    return response.json()

def rtr_command_status(self, cloud_request_id: str, sequence_id: int):
    """
    - cloud_request_id (string) (required) - the request ID from execute_command()
    - sequence_id (integer) (required) - command responses are chunked across sequences

    returns a JSON blob
    """
    uri = '/real-time-response/entities/command/v1'
    method = 'get'

    kwargs = {
        'cloud_request_id' : cloud_request_id,
        'sequence_id' : sequence_id,
    }
    args_validation = {
        'cloud_request_id' : str,
        'sequence_id' : int,
    }
    validate_kwargs(args_validation, kwargs, required=args_validation.keys())

    response = self.request(uri=uri,
                            request_method=method,
                            data=kwargs,
                            )
    logger.debug(f"Request body: {response.request.body}")
    logger.debug(f"Response: {response.text}")

    return response.json()

def rtr_command_status_wait(self, cloud_request_id: str, interval: int = 5, maxtime: int = 60):
    """ polls rtr_command_status() periodically until it's done, then returns the results

        - interval (int) the number of seconds between a check
        - maxtime (int) the maximum time to wait - note this will currently just calculate maxtime / interval, so isn't exact (yet)
    """
    finished = False
    iteration = 0
    start_time = int(time.time())
    while not finished:
        response = self.rtr_command_status(cloud_request_id=cloud_request_id,
                                           sequence_id=0,
                                           )
        logger.debug(response)

        if response.get('resources', [{}])[0].get('complete'):
            return response
        iteration += 1

        if (int(time.time()) - start_time) >= maxtime:
            raise TimeoutError(f"Waited too long for {cloud_request_id} to finish, last response: {json.dumps(response)}")
        logger.debug(f"Sleeping for {interval} second(s)")
        time.sleep(interval)
    return response



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['crowdstrike']

package_data = \
{'': ['*']}

install_requires = \
['loguru>=0.6.0,<0.7.0', 'requests-oauthlib>=1.3.1,<2.0.0']

setup_kwargs = {
    'name': 'crowdstrike',
    'version': '0.0.5',
    'description': 'Python interface to the Crowdstrike API',
    'long_description': '* THIS IS NO LONGER UPDATED *\n\n\n# Crowdstrike API\n\nImplements some of the functions to interface with the [Crowdstrike APIs](https://assets.falcon.crowdstrike.com/support/api/swagger.html).\n\nWant to contribute? Log an issue or PR on the Repo.\n\nTo enable logging, use [loguru](https://github.com/Delgan/loguru) and run `logger.enable("crowdstrike")` in your script.\n\n## Checking that all the endpoints are covered\n\n`validate_api_endpoints.py` needs the `swagger.json` file from the documentation page on [crowdstrike.com](https://assets.falcon.crowdstrike.com/support/api/swagger.html), then you can check everything has an actionable method.\n\nEg:\n\n    2020-10-14 18:56:57.801 | INFO     | __main__:<module>:60 - [OK] create_rtr_session() implements /real-time-response/entities/sessions/v1 : post\n    2020-10-14 18:56:57.801 | INFO     | __main__:<module>:60 - [OK] delete_rtr_session() implements /real-time-response/entities/sessions/v1 : delete\n    2020-10-14 18:56:57.802 | ERROR    | __main__:<module>:64 - Path not found /real-time-response/queries/put-files/v1 : get\n    2020-10-14 18:56:57.802 | INFO     | __main__:<module>:60 - [OK] search_rtr_scripts() implements /real-time-response/queries/scripts/v1 : get\n    2020-10-14 18:56:57.802 | INFO     | __main__:<module>:60 - [OK] list_rtr_session_ids() implements /real-time-response/queries/sessions/v1 : get\n    2020-10-14 18:56:57.802 | ERROR    | __main__:<module>:64 - Path not found /samples/entities/samples/v2 : post\n    2020-10-14 18:56:57.802 | ERROR    | __main__:<module>:64 - Path not found /sensors/combined/installers/v1 : get\n    2020-10-14 18:56:57.802 | ERROR    | __main__:<module>:64 - Path not found /sensors/entities/datafeed-actions/v1/{partition} : post\n    2020-10-14 18:56:57.803 | ERROR    | __main__:<module>:64 - Path not found /sensors/entities/datafeed/v2 : get\n    2020-10-14 18:56:57.803 | ERROR    | __main__:<module>:64 - Path not found /sensors/entities/download-installer/v1 : get\n\n... lots to do.',
    'author': 'James Hodgkinson',
    'author_email': 'james@terminaloutcomes.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/yaleman/crowdstrike_api',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""
Crowdstrike API module

"""

import json
import os
import sys
import time
import errno

try:
    from loguru import logger
    import requests
    from requests.auth import HTTPBasicAuth

    import requests_oauthlib
    from oauthlib.oauth2 import BackendApplicationClient, TokenExpiredError
except ImportError as importerror:
    sys.exit(f"Failed to import a dependency, quitting. Error: {importerror}")

from .hosts import hosts_query_devices, host_action, hosts_hidden, hosts_detail
from .hostgroup import search_host_groups, get_host_groups, update_host_group, create_host_group, delete_host_groups
from .sensor_download import get_sensor_installer_details, get_ccid, get_latest_sensor_id, get_sensor_installer_ids, download_sensor
from .event_streams import get_event_streams
from .incidents import incidents_behaviors_by_id, incidents_get_crowdscores, incidents_get_details, incidents_perform_actions, incidents_query, incidents_query_behaviors
from .detects import get_detects, get_detections
from .rtr import create_rtr_session, delete_rtr_session, rtr_execute_command, rtr_command_status, rtr_command_status_wait, list_rtr_session_ids
from .rtr_admin import search_rtr_scripts, get_rtr_scripts

from .intel import *
try:
    if isinstance(THIS_SHOULD_FAIL, bool):
        raise ValueError("__all__ is missing in a module")
except ValueError:
    logger.error("__all__ is missing in a module")
    sys.exit(1)
except NameError:
    # this is good
    pass

API_BASEURL = "https://api.crowdstrike.com"

# if you want to enable logging then you can just run logger.enable("crowdstrike") in your code.
logger.disable('crowdstrike')

# set status by using the keys in here, not the values
INCIDENT_STATUS_LOOKUP = {
    20 : "New",
    25 : "Reopened",
    30 : "In progress",
    40 : "Closed",
}

# normal ones they use
INCIDENT_TAGS_STANDARD = [
    'False Positive',
    'Ignored',
    'Lateral Movement',
    'Testing',
    'True Positive',
]


class CrowdstrikeAPI:
    """ Crowdstrike API """
    def __init__(self,
                 client_id: str,
                 client_secret: str,
                 api_baseurl: str = API_BASEURL):
        """ Starts up the CrowdstrikeAPI module Needs two strings,
        the client_id and client_secret, available from
        https://falcon.crowdstrike.com/support/api-clients-and-keys

        You might want to override api_baseurl if you're in a different region, too.
        """
        self.client_id = client_id
        self.client_secret = client_secret
        self.client = BackendApplicationClient(client_id=client_id)
        self.oauth = requests_oauthlib.OAuth2Session(client=self.client)
        #grab a token to start with
        self.token = self.get_token()
        self.api_baseurl = api_baseurl

    def get_token(self):
        """ Gets the latest auth token and returns it. """
        logger.debug("Requesting auth token")
        uri = '/oauth2/token'
        # this is just here for introspection purposes
        # pylint: disable=unused-variable
        method = 'post'
        self.token = self.oauth.fetch_token(
            token_url=f"{API_BASEURL}{uri}",
            client_id=self.client_id,
            client_secret=self.client_secret,
        )
        return self.token

    def revoke_token(self, **kwargs):
        """ revoke the current token

        https://assets.falcon.crowdstrike.com/support/api/swagger.html#/oauth2/oauth2RevokeToken
        """
        #token = kwargs.get('token', self.token)
        #logger.debug(self.token)
        token = kwargs.get('token', self.token.get('access_token'))

        uri = '/oauth2/revoke'
        method = 'post'
        data = {
            'token': token,
        }
        response = requests.request(url=f"{API_BASEURL}{uri}",
                                    method=method,
                                    data=data,
                                    auth=HTTPBasicAuth(self.client_id, self.client_secret)
                                    )
        logger.debug(response)
        logger.debug(response.request.headers)
        logger.debug(response.request.body)
        return response.json()

    def do_request(self, uri: str, data: dict, request_method: str = 'get'):
        """ does the request, this allows a single code implementation for
            the duplicated calls in self.request()

            default request method is get
        """
        fulluri = f"{API_BASEURL}{uri}"

        # these methods use a get-request-style-data-in-the-url nastiness.
        methods_using_params = ['get', 'delete']

        if request_method.lower() in methods_using_params and data:
            response = self.oauth.request(request_method, fulluri, params=data)
        else:
            response = self.oauth.request(request_method, fulluri, json=data)
        return response

    def request(self, uri: str, request_method: str = None, data: dict = None):
        """ does a request

        request_method is a string, either get / post / delete etc
            default is set in self.do_request()
        """
        # TODO: handle rate limiting
        # Requests will return the following headers:
        # X-RateLimit-Limit : Request limit per minute.
        #   type = integer
        # X-RateLimit-Remaining : The number of requests remaining for the sliding 1 minute window.
        #   type = integer
        logger.debug(f"request(uri='{uri}', request_method='{request_method}', data='{data}'")
        try:
            req = self.do_request(uri=uri,
                                  request_method=request_method,
                                  data=data,
                                  )
        except TokenExpiredError:
            logger.debug("Token's expired, grabbing a new one")
            self.token = self.get_token()
            req = self.do_request(uri=uri,
                                  request_method=request_method,
                                  data=data,
                                 )
            req.raise_for_status()
        return req

    # sensor-related things
    get_ccid = get_ccid
    get_latest_sensor_id = get_latest_sensor_id
    get_sensor_installer_details = get_sensor_installer_details
    get_sensor_installer_ids = get_sensor_installer_ids
    download_sensor = download_sensor
    #detects
    get_detects = get_detects
    get_detections = get_detections
    # event-streams
    get_event_streams = get_event_streams
    # hosts
    hosts_query_devices = hosts_query_devices
    host_action = host_action
    hosts_hidden = hosts_hidden
    hosts_detail = hosts_detail

    # incidents
    incidents_behaviors_by_id = incidents_behaviors_by_id
    incidents_get_crowdscores = incidents_get_crowdscores
    incidents_get_details = incidents_get_details
    incidents_perform_actions = incidents_perform_actions
    incidents_query = incidents_query
    incidents_query_behaviors = incidents_query_behaviors

    #hostgroups
    create_host_group = create_host_group
    search_host_groups = search_host_groups
    get_host_groups = get_host_groups
    update_host_group = update_host_group
    delete_host_groups = delete_host_groups

    # intel
    get_intel_indicators = get_intel_indicators

    #rtr
    create_rtr_session = create_rtr_session
    delete_rtr_session = delete_rtr_session
    list_rtr_session_ids = list_rtr_session_ids

    rtr_command_status = rtr_command_status
    rtr_execute_command = rtr_execute_command
    rtr_command_status_wait = rtr_command_status_wait

    #rtr_admin
    search_rtr_scripts = search_rtr_scripts
    get_rtr_scripts = get_rtr_scripts



""" utility functions for the crowdstrike API """

def validate_kwargs(args_validation: dict, kwargs: dict, required: list = None):
    """ validates arguments pushed to the function

    args
    - args_validation: dict of key/types to check, example below checks if action is a string and ids is a list
            args_validation = {
                'action' : str,
                'ids' : list,
            }
    - required: if passed, it'll require those keys to be set

    """
    for key in kwargs:
        if key not in args_validation:
            raise ValueError(f"{key} not a valid argument")
        if not isinstance(kwargs[key], args_validation[key]):
            raise TypeError(f"{key} not the valid type, should be: {args_validation[key]}, was {type(kwargs[key])}")
    if required:
        for key in required:
            if key not in kwargs:
                raise ValueError(f"argument {key} needs to be set")
    return True



""" handles interactions with the "hosts" category """

# import json

from loguru import logger

from .utilities import validate_kwargs

def search_rtr_scripts(self, **kwargs):
    """ Search for Realtime Response scripts in your environment by providing an FQL filter and paging details.
    Returns a set of Host Groups which match the filter criteria

    - filter (string) - The filter expression that should be used to limit the results
    - offset (integer) - The offset to start retrieving records from
    - limit (integer) - The maximum records to return. [1-5000]
    - sort (string) - field to sort by
    """
    args_validation = {
        'filter' : str,
        'offset' : int,
        'limit' : int,
        'sort' : str,

    }
    validate_kwargs(args_validation, kwargs)

    uri = '/real-time-response/queries/scripts/v1'
    method = 'get'

    response = self.request(uri=uri,
                            request_method=method,
                            data=kwargs,
                            )
    logger.debug(f"Request body: {response.request.body}")
    return response.json()

def get_rtr_scripts(self, **kwargs):
    """ Get custom-scripts based on the IDs given, tehse are used for the RTR `runscript` command

    - ids (list) - A list of file IDs for the scripts to show
    """
    args_validation = {
        'ids' : list,
    }
    validate_kwargs(args_validation, kwargs, required=args_validation.keys())

    uri = '/real-time-response/entities/scripts/v1'
    method = 'get'

    response = self.request(uri=uri,
                            request_method=method,
                            data=kwargs,
                            )
    logger.debug(f"Request body: {response.request.body}")
    return response.json()


def execute_rtr_admin_command(self, **kwargs): #pylint: disable=unused-argument
    """
        - cloud_request_id (string) (required) - Cloud Request ID of the executed command to query
        - sequence_id (integer) (required) - Sequence ID that we want to retrieve. Command responses are chunked across sequences
    """

    args_validation = {
        'cloud_request_id' : str,
        'sequence_id' : int,

    }
    validate_kwargs(args_validation, kwargs, required=kwargs.keys())
    uri = '/real-time-response/entities/admin-command/v1'
    method = 'get'
    response = self.request(uri=uri,
                            request_method=method,
                            data=kwargs,
                            )
    logger.debug(f"Request body: {response.request.body}")
    return response.json()

def status_rtr_admin_command(self, **kwargs):
    """
    Use this endpoint to run these real time response commands:
        cat         cd
        clear        cp
        encrypt        env
        eventlog        filehash
        get        getsid
        help        history
        ipconfig        kill
        ls        map
        memdump        mkdir
        mount        mv
        netstat        ps
        put        reg query
        reg set        reg delete
        reg load        reg unload
        restart        rm
        run        runscript
        shutdown        unmap
        xmemdump        zip
    - base_command (string) Active-Responder command type we are going to execute, for example: get or cp. Refer to the RTR documentation for the full list of commands.
    - command_string (string) Full command string for the command. For example get some_file.txt
    - session_id (string) RTR session ID to run the command on

    Example Value
        Model
        {
        "base_command": "cat",
        "command_string": "cat c:\temp\badfile.txt",
        "session_id": "string"
        }
    """
    args_validation = {
        'body' : str,
        'command_string' : str,
        'session_id' : str,
    }
    validate_kwargs(args_validation, kwargs, required=kwargs.keys())
    uri = '/real-time-response/entities/admin-command/v1'
    method = 'post'
    response = self.request(uri=uri,
                            request_method=method,
                            data=kwargs,
                            )
    logger.debug(f"Request body: {response.request.body}")
    return response.json()
